#include "server_data_interact.h"
#include <engine_engine_number_9/baza_throw.h>

namespace iu9cawebchat {
    int64_t get_current_history_id_of_chat(SqliteConnection& conn, int64_t chatId) {
        SqliteStatement req(conn, "SELECT `it_HistoryId` FROM `chat` WHERE `id` = ?1", {{1, chatId}}, {});
        fsql_integer_or_null HistoryId;
        int status = sqlite_stmt_step(req, {{0, &HistoryId}}, {});
        een9_ASSERT_pl(status == SQLITE_ROW);
        return HistoryId.value;
    }

    int64_t get_current_history_id_of_user_chatList(SqliteConnection& conn, int64_t userId) {
        SqliteStatement req(conn, "SELECT `chatList_HistoryId` FROM `user` WHERE `id` = ?1", {{1, userId}}, {});
        fsql_integer_or_null HistoryId;
        int status = sqlite_stmt_step(req, {{0, &HistoryId}}, {});
        een9_ASSERT_pl(status == SQLITE_ROW);
        return HistoryId.value;
    }

    void internalapi_pollEvents_in_chat_collect_membership_events(SqliteConnection& conn,
        std::vector<json::JSON>& events, int64_t chatId, int64_t LocalHistoryId) {
        SqliteStatement membership_changes(conn,
                    "SELECT `userId`, `role`, FROM `user_chat_membership` WHERE `chatId` = ?1 "
                    "AND `chat_IncHistoryId` > ?2", {{1, chatId}, {2, LocalHistoryId}}, {});
        fsql_integer_or_null ev_userId;
        fsql_integer_or_null ev_user_role;
        while (true) {
            int status = sqlite_stmt_step(membership_changes,
            {{0, &ev_userId}, {1, &ev_user_role}}, {});
            if (status != SQLITE_ROW)
                break;
            events.emplace_back();
            json::JSON& event = events.back();
            event["member"] = json::JSON(ev_userId.value);
            if (ev_user_role.value == user_chat_role_deleted) {
                event["type"] = json::JSON("removedMember");
            } else {
                event["type"] = json::JSON("addedMember");
                RowUser_Content USER = lookup_user_content(conn, ev_userId.value);
                event["content"]["name"] = json::JSON(USER.name);
                event["content"]["nickname"] = json::JSON(USER.nickname);
                event["content"]["role"] = json::JSON(stringify_user_chat_role(ev_user_role.value));
            }
            events.push_back(event);
        }
    }

    void internalapi_pollEvents_in_chat_collect_messages_events(SqliteConnection& conn,
        std::vector<json::JSON>& events, int64_t chatId, int64_t LocalHistoryId) {
        SqliteStatement messages_changes(conn,
            "SELECT `id`, `previous`, `senderUserId`, `exists`, `isSystem`, `text` FROM `messages` WHERE "
            "WHERE `chatId` = ?1 AND `chat_IncHistoryId` > ?2", {{1, chatId}, {2, LocalHistoryId}}, {});
        fsql_integer_or_null ev_msgId, ev_previousMsgId, msgSenderUserId, msgExists, msgIsSystem;
        fsql_text8_or_null msgText;
        while (true) {
            int status = sqlite_stmt_step(messages_changes,
                {{0, &ev_msgId}, {1, &ev_previousMsgId}, {2, &msgSenderUserId}, {3, &msgExists}, {4, &msgIsSystem}},
                {{5, &msgText}});
            if (status != SQLITE_ROW)
                break;
            events.emplace_back();
            json::JSON& event = events.back();
            event["type"] = json::JSON("newMessage");
            event["id"] = json::JSON(ev_msgId.value);
            event["previous"] = json::JSON(ev_previousMsgId.value);
            event["content"]["sender"] = json::JSON(msgSenderUserId.value);
            event["content"]["isSystem"] = json::JSON((bool)msgIsSystem.value);
            event["content"]["text"] = json::JSON(msgText.value);
            events.push_back(event);
        }
    }

    void internalapi_pollEvents_in_user_chatList_collect_events(SqliteConnection& conn,
        std::vector<json::JSON>& events, int64_t userId, int64_t LocalHistoryId) {
        SqliteStatement membership_changes(conn,
            "SELECT `chatId`, `role` FROM `user_chat_membership` WHERE `userId` = ?1 "
            "AND `user_chatList_IncHistoryId` > ?2", {{1, userId}, {2, LocalHistoryId}});
        fsql_integer_or_null ev_chatId, usersRoleHere;
        while (true) {
            int status = sqlite_stmt_step(membership_changes, {{0, &ev_chatId}, {1, &usersRoleHere}}, {});
            if (status != SQLITE_ROW)
                break;
            events.emplace_back();
            json::JSON& event = events.back();
            event["id"] = json::JSON(ev_chatId.value);
            if (usersRoleHere.value == user_chat_role_deleted) {
                event["type"] = json::JSON("removedChat");
            } else {
                event["type"] = json::JSON("addedChat");
                RowChat_Content CHAT = lookup_chat_content(conn, ev_chatId.value);
                event["content"]["name"] = json::JSON(CHAT.name);
                event["content"]["nickname"] = json::JSON(CHAT.nickname);
                event["content"]["lastMsgId"] = json::JSON(CHAT.lastMsgId);
                event["content"]["roleHere"] = json::JSON(stringify_user_chat_role(usersRoleHere.value));
            }
        }
    }

    json::JSON internalapi_pollEvents(SqliteConnection& conn, int64_t uid, const json::JSON& Sent) {
        json::JSON Recv;
        Recv["status"] = json::JSON(0l);
        Recv["update"] = json::JSON(json::array);
        const std::vector<json::JSON>& req_scope = Sent["scope"].g().asArray();
        std::vector<json::JSON>& updated = Recv["update"].g().asArray();
        for (const json::JSON& hist_entity_request: req_scope) {
            updated.emplace_back();
            json::JSON& hist_entity_response = updated.back();
            hist_entity_response["type"] = hist_entity_request["type"].g();
            hist_entity_response["events"] = json::JSON(json::array);
            std::vector<json::JSON>& events = hist_entity_response["events"].g().asArray();
            const int64_t LocalHistoryId = hist_entity_request["LocalHistoryId"].g().asInteger().get_int();
            if (hist_entity_request["type"].g().asString() == "chat") {
                int64_t chatId = hist_entity_request["chatId"].g().asInteger().get_int();
                if (get_role_of_user_in_chat(conn, uid, chatId) == user_chat_role_deleted)
                    een9_THROW("internalapi/pollEvents: trying to access chat that user does not belong to");
                hist_entity_response["chatId"] = json::JSON(chatId);
                hist_entity_response["HistoryId"] = json::JSON(get_current_history_id_of_chat(conn, chatId));
                /* Two classes of 'real events' can happen to chat: membership table change, message table change */
                /* Here, I collect membership changes (related to this chat) */
                internalapi_pollEvents_in_chat_collect_membership_events(conn, events, chatId, LocalHistoryId);
                /* Here, I collect message changes (related to this chat) */
                internalapi_pollEvents_in_chat_collect_messages_events(conn, events, chatId, LocalHistoryId);
            } else if (hist_entity_request["type"].g().asString() == "chatlist") {
                hist_entity_response["HistotyId"] = json::JSON(get_current_history_id_of_user_chatList(conn, uid));
                internalapi_pollEvents_in_user_chatList_collect_events(conn, events, uid, LocalHistoryId);
            } else
                een9_THROW("Bad request");
        }
        return Recv;
    }
}
